<footer>
  {% if (theme_prev_next_buttons_location == 'bottom' or theme_prev_next_buttons_location == 'both') and (next or prev) %}
    <div class="rst-footer-buttons" role="navigation" aria-label="footer navigation">
      {% if next %}
        <a href="{{ next.link|e }}" class="btn btn-neutral float-right" title="{{ next.title|striptags|e }}" accesskey="n" rel="next">{{ _('Next') }}
          <img src="{{ pathto('_static/images/chevron-right-orange.svg', 1) }}" alt="right arrow" class="next-page">
        </a>
      {% endif %}
      {% if prev %}
        <a href="{{ prev.link|e }}" class="btn btn-neutral" title="{{ prev.title|striptags|e }}" accesskey="p" rel="prev">
          <img src="{{ pathto('_static/images/chevron-right-orange.svg', 1) }}" alt="left arrow" class="previous-page"> {{ _('Previous') }}
        </a>
      {% endif %}
    </div>
  {% endif %}

  {% if theme_pytorch_project == 'tutorials' %}

    <hr class="rating-hr hr-top">
      <div class="rating-container">
        <div class="rating-prompt">Rate this Tutorial</div>
        <div class="stars-outer">
          <i class="far fa-star" title="1 Star" data-behavior="tutorial-rating" data-count="1"></i>
          <i class="far fa-star" title="2 Stars" data-behavior="tutorial-rating" data-count="2"></i>
          <i class="far fa-star" title="3 Stars" data-behavior="tutorial-rating" data-count="3"></i>
          <i class="far fa-star" title="4 Stars" data-behavior="tutorial-rating" data-count="4"></i>
          <i class="far fa-star" title="5 Stars" data-behavior="tutorial-rating" data-count="5"></i>
        </div>
      </div>
    <hr class="rating-hr hr-bottom"/>

  {% else %}

    <hr>

  {% endif %}

  <div role="contentinfo">
    <table>
      <tr>
        <td>
            <p>
            {%- if show_copyright %}
              {%- if hasdoc('copyright') %}
                {% trans path=pathto('copyright'), copyright=copyright|e %}&copy; <a href="{{ path }}">Copyright</a> {{ copyright }}.{% endtrans %}
              {%- else %}
                {% trans copyright=copyright|e %}&copy; Copyright {{ copyright }}.{% endtrans %}
              {%- endif %}
            {%- endif %}

            {%- if build_id and build_url %}
              {% trans build_url=build_url, build_id=build_id %}
                <span class="build">
                  Build
                  <a href="{{ build_url }}">{{ build_id }}</a>.
                </span>
              {% endtrans %}
            {%- elif commit %}
              {% trans commit=commit %}
                <span class="commit">
                  Revision <code>{{ commit }}</code>.
                </span>
              {% endtrans %}
            {%- elif last_updated %}
              {% trans last_updated=last_updated|e %}Last updated on {{ last_updated }}.{% endtrans %}
            {%- endif %}
            </p>

          {%- if show_sphinx %}
            {% trans %}
              <div>
                Built with <a href="http://sphinx-doc.org/">Sphinx</a> using a <a href="https://github.com/rtfd/sphinx_rtd_theme">theme</a> provided by <a href="https://readthedocs.org">Read the Docs</a>.
              </div>
          {% endtrans %}
        {%- endif %}
        </td>
        <td>
          {%- if show_netlify_badge %}
            {% trans %}
              <div>
                <a href="https://www.netlify.com" target="_blank" rel="noopener noreferrer">
                  <img
                  src="_static/img/netlify-light.svg"
                  alt="Deploys by Netlify"
                  width=114
                  height=51 />
                </a>
              </div>
            {% endtrans %}
          {%- endif %}
        </td>
      </tr>
    </table>
  </div>

  {%- block extrafooter %} {% endblock %}

</footer>
